import numpy as np
import xtrack as xt
import xdeps as xd

def expr_to_mad_str(expr):

    expr_str = str(expr)

    fff = xt.line.Functions()
    for nn in fff._mathfunctions:
        expr_str = expr_str.replace(f'f.{nn}(', f'{nn}(')

    expr_str = expr_str.replace("'", "")
    expr_str = expr_str.replace('"', "")

    # transform vars[...] in (...)
    while "vars[" in expr_str:
        before, after = tuple(*[expr_str.split("vars[", 1)])
        # find the corresponding closing bracket
        count = 1
        for ii, cc in enumerate(after):
            if cc == "]":
                count -= 1
            elif cc == "[":
                count += 1
            if count == 0:
                break

        expr_str = before + "(" + after[:ii] + ")" + after[ii+1:]

        expr_str = expr_str.replace("**", "^")

    return expr_str

def mad_str_or_value(var):
    if _is_ref(var):
        out = expr_to_mad_str(var)
        out = out.strip('._expr')
        return out
    else:
        return var

def mad_assignment(lhs, rhs):
    if _is_ref(rhs):
        rhs = mad_str_or_value(rhs)
    if isinstance(rhs, str):
        return f"{lhs} := {rhs}"
    else:
        return f"{lhs} = {rhs}"

_ge = xt.elements._get_expr
_is_ref = xd.refs.is_ref

def _knl_ksl_to_mad(mult):
    knl_mad = []
    ksl_mad = []
    for kl, klmad in zip([mult.knl, mult.ksl], [knl_mad, ksl_mad]):
        for ii in range(len(kl._value)):
            item = mad_str_or_value(_ge(kl[ii]))
            if not isinstance(item, str):
                item = str(item)
            klmad.append(item)
    knl_token = 'knl:={' + ','.join(knl_mad) + '}'
    ksl_token = 'ksl:={' + ','.join(ksl_mad) + '}'
    return knl_token, ksl_token

def _get_eref(line, name):
    return line.element_refs[name]

def _handle_transforms(tokens, el):
    if el.shift_x._expr is not None or el.shift_x._value:
        tokens.append(mad_assignment('dx', _ge(el.shift_x)))
    if el.shift_y._expr is not None or el.shift_y._value:
        tokens.append(mad_assignment('dy', _ge(el.shift_y)))
    if el.rot_s_rad._expr is not None or el.rot_s_rad._value:
        tokens.append(mad_assignment('tilt', _ge(el.rot_s_rad)))
    if el.shift_s._expr is not None or el.shift_s._value:
        raise NotImplementedError("shift_s is not yet supported in mad writer")

def cavity_to_madx_str(name, line):
    cav = _get_eref(line, name)
    tokens = []
    tokens.append('rfcavity')
    tokens.append(mad_assignment('freq', _ge(cav.frequency) * 1e-6))
    tokens.append(mad_assignment('volt', _ge(cav.voltage) * 1e-6))
    tokens.append(mad_assignment('lag', _ge(cav.lag) / 360.))
    _handle_transforms(tokens, cav)

    return ', '.join(tokens)

def marker_to_madx_str(name, line):
    if name.endswith('_entry'):
         parent_name = name.replace('_entry', '')
         if (parent_name in line.element_dict):
             return None
    if name.endswith('_exit'):
        parent_name = name.replace('_exit', '')
        if (parent_name in line.element_dict):
            return None
    return 'marker'

def drift_to_madx_str(name, line):
    drift = _get_eref(line, name)
    tokens = []
    tokens.append('drift')
    tokens.append(mad_assignment('l', _ge(drift.length)))
    return ', '.join(tokens)

def multipole_to_madx_str(name, line):
    mult = _get_eref(line, name)

    if (len(mult.knl._value) == 1 and len(mult.ksl._value) == 1
        and mult.hxl._value == 0):
        # It is a dipole corrector
        tokens = []
        tokens.append('kicker')
        tokens.append(mad_assignment('hkick', -1 * _ge(mult.knl[0])))
        tokens.append(mad_assignment('vkick', _ge(mult.ksl[0])))
        tokens.append(mad_assignment('lrad', _ge(mult.length)))

        _handle_transforms(tokens, mult)

        return ', '.join(tokens)

    # correctors are not handled correctly!!!!
    # https://github.com/MethodicalAcceleratorDesign/MAD-X/issues/911
    # assert mult.hyl._value == 0

    tokens = []
    tokens.append('multipole')
    knl_token, ksl_token = _knl_ksl_to_mad(mult)
    tokens.append(knl_token)
    tokens.append(ksl_token)
    tokens.append(mad_assignment('lrad', _ge(mult.length)))
    tokens.append(mad_assignment('angle', _ge(mult.hxl)))

    _handle_transforms(tokens, mult)

    return ', '.join(tokens)

def rfmultipole_to_madx_str(name, line):
    rfmult = _get_eref(line, name)

    tokens = []
    tokens.append('rfmultipole')
    knl_mad = []
    ksl_mad = []
    for kl, klmad in zip([rfmult.knl, rfmult.ksl], [knl_mad, ksl_mad]):
        for ii in range(len(kl._value)):
            item = mad_str_or_value(_ge(kl[ii]))
            if not isinstance(item, str):
                item = str(item)
            klmad.append(item)
    pnl_mad = []
    psl_mad = []
    for pp, plmad in zip([rfmult.pn, rfmult.ps], [pnl_mad, psl_mad]):
        for ii in range(len(pp._value)):
            item = mad_str_or_value(_ge(pp[ii]) / 360)
            if not isinstance(item, str):
                item = str(item)
            plmad.append(item)

    tokens.append('knl:={' + ','.join(knl_mad) + '}')
    tokens.append('ksl:={' + ','.join(ksl_mad) + '}')
    tokens.append('pnl:={' + ','.join(pnl_mad) + '}')
    tokens.append('psl:={' + ','.join(psl_mad) + '}')
    tokens.append(mad_assignment('freq', _ge(rfmult.frequency) * 1e-6))
    tokens.append(mad_assignment('volt', _ge(rfmult.voltage) * 1e-6))
    tokens.append(mad_assignment('lag', _ge(rfmult.lag) / 360.))

    _handle_transforms(tokens, rfmult)

    return ', '.join(tokens)

def dipoleedge_to_madx_str(name, line):
    raise NotImplementedError("isolated dipole edges are not yet supported")

def bend_to_madx_str(name, line, bend_type='sbend'):

    assert bend_type in ['sbend', 'rbend']

    bend = _get_eref(line, name)

    tokens = []
    tokens.append(bend_type)
    tokens.append(mad_assignment('l', _ge(bend.length)))
    tokens.append(mad_assignment('angle', _ge(bend.h) * _ge(bend.length)))
    tokens.append(mad_assignment('k0', _ge(bend.k0)))
    tokens.append(mad_assignment('e1', _ge(bend.edge_entry_angle)))
    tokens.append(mad_assignment('e2', _ge(bend.edge_exit_angle)))
    tokens.append(mad_assignment('fint', _ge(bend.edge_entry_fint)))
    tokens.append(mad_assignment('fintx', _ge(bend.edge_exit_fint)))
    tokens.append(mad_assignment('hgap', _ge(bend.edge_entry_hgap)))
    tokens.append(mad_assignment('k1', _ge(bend.k1)))
    knl_token, ksl_token = _knl_ksl_to_mad(bend)
    tokens.append(knl_token)
    tokens.append(ksl_token)

    _handle_transforms(tokens, bend)

    return ', '.join(tokens)

def rbend_to_madx_str(name, line):
    return bend_to_madx_str(name, line, bend_type='rbend')

def sextupole_to_madx_str(name, line):
    sext = _get_eref(line, name)
    tokens = []
    tokens.append('sextupole')
    tokens.append(mad_assignment('l', _ge(sext.length)))
    tokens.append(mad_assignment('k2', _ge(sext.k2)))
    tokens.append(mad_assignment('k2s', _ge(sext.k2s)))

    _handle_transforms(tokens, sext)

    return ', '.join(tokens)

def octupole_to_madx_str(name, line):
    octup = _get_eref(line, name)
    tokens = []
    tokens.append('octupole')
    tokens.append(mad_assignment('l', _ge(octup.length)))
    tokens.append(mad_assignment('k3', _ge(octup.k3)))
    tokens.append(mad_assignment('k3s', _ge(octup.k3s)))

    _handle_transforms(tokens, octup)

    return ', '.join(tokens)

def quadrupole_to_madx_str(name, line):
    quad = _get_eref(line, name)
    tokens = []
    tokens.append('quadrupole')
    tokens.append(mad_assignment('l', _ge(quad.length)))
    tokens.append(mad_assignment('k1', _ge(quad.k1)))
    tokens.append(mad_assignment('k1s', _ge(quad.k1s)))
    knl_token, ksl_token = _knl_ksl_to_mad(quad)
    tokens.append(knl_token)
    tokens.append(ksl_token)

    _handle_transforms(tokens, quad)

    return ', '.join(tokens)

def solenoid_to_madx_str(name, line):
    sol = _get_eref(line, name)
    tokens = []
    tokens.append('solenoid')
    tokens.append(mad_assignment('l', _ge(sol.length)))
    tokens.append(mad_assignment('ks', _ge(sol.ks)))
    tokens.append(mad_assignment('ksi', _ge(sol.ksi)))

    _handle_transforms(tokens, sol)

    return ', '.join(tokens)

def srotation_to_madx_str(name, line):
    raise NotImplementedError("isolated rotations are not yet supported")
    return 'marker'
    # srot = _get_eref(line, name)
    # tokens = []
    # tokens.append('srotation')
    # tokens.append(mad_assignment('angle', _ge(srot.angle)*np.pi/180.))
    # return ', '.join(tokens)

xsuite_to_mad_conveters={
    xt.Cavity: cavity_to_madx_str,
    xt.Marker: marker_to_madx_str,
    xt.Drift: drift_to_madx_str,
    xt.Multipole: multipole_to_madx_str,
    xt.DipoleEdge: dipoleedge_to_madx_str,
    xt.Bend: bend_to_madx_str,
    xt.RBend: rbend_to_madx_str,
    xt.Sextupole: sextupole_to_madx_str,
    xt.Octupole: octupole_to_madx_str,
    xt.Quadrupole: quadrupole_to_madx_str,
    xt.Solenoid: solenoid_to_madx_str,
    xt.SRotation: srotation_to_madx_str,
    xt.RFMultipole: rfmultipole_to_madx_str,
}

def to_madx_sequence(line, name='seq', mode='sequence'):
    # build variables part
    vars_str = ""
    for vv in line.vars.keys():
        if vv == '__vary_default':
            continue
        vars_str += mad_assignment(vv, _ge(line.vars[vv])) + ";\n"

    if mode =='line':
        elements_str = ""
        for nn in line.element_names:
            el = line[nn]
            el_str = xsuite_to_mad_conveters[el.__class__](nn, line)
            elements_str += f"{nn}: {el_str};\n"
        line_str = f'{name}: line=(' + ', '.join(line.element_names) + ');'
        machine_str = elements_str + line_str
    elif mode == 'sequence':
        tt = line.get_table()
        line_length = tt['s', -1]
        seq_str = f'{name}: sequence, l={line_length};\n' #, refer=entry;\n'
        # s_dict = {nn:ss for nn, ss in zip(tt.name, tt.s)}

        s_dict = {}
        tt_name = tt.name
        tt_s = tt.s
        tt_isthick = tt.isthick
        for ii in range(len(tt.name)):
            nn = tt_name[ii]
            if not(tt_isthick[ii]):
                s_dict[nn] = tt_s[ii]
            else:
                s_dict[nn] = 0.5 * (tt_s[ii] + tt_s[ii+1])

        for nn in line.element_names:
            el = line.element_dict[nn]
            el_str = xsuite_to_mad_conveters[el.__class__](nn, line)
            if nn + '_tilt_entry' in line.element_dict:
                el_str += ", " + mad_assignment('tilt',
                            _ge(line.element_refs[nn + '_tilt_entry'].angle) / 180. * np.pi)

            if el_str is None:
                continue

            nn_mad = nn.replace(':', '__') # : not supported in madx names
            seq_str += f"{nn_mad}: {el_str}, at={s_dict[nn]};\n"
        seq_str += 'endsequence;'
        machine_str = seq_str

    mad_input = vars_str + '\n' + machine_str + '\n'
    return mad_input


import numpy as np
import xtrack as xt

# Save to json
line = xt.Line.from_json('toy_ring.json')
tw0 = line.twiss(method='4d')

xtsl = xt.slicing
# slicing to see beta inside bends
line_sliced = line.copy()
line_sliced.slice_thick_elements(
    slicing_strategies=[
        xtsl.Strategy(slicing=None), # default
        xtsl.Strategy(slicing=xtsl.Teapot(100, mode='thick'), element_type=xt.Bend),
        xtsl.Strategy(slicing=xtsl.Teapot(100, mode='thick'), element_type=xt.Quadrupole),
    ]
)

# tw = line.twiss(
#     start='mqf.1', end=len(line)-1, init=xt.TwissInit(betx=1, bety=1))
tw = line_sliced.twiss(method='4d')

tt = line.get_table(attr=True)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1, figsize=(6.4, 4.8 * 1.4))

ax0 = plt.subplot(3, 1, 1)
ax0.bar(tt.s, tt.k1l, width=tt.length, color='r', alpha=0.2)
ax0.axhline(0, color='k', lw=0.5, alpha=0.5)
ax0.set_ylabel(r'k1L [m$^{-1}$]')
ax0.set_ylim(np.array([-1, 1]) * np.max(np.abs(tt.k1l)) * 1.1)

ax1 = plt.subplot(3, 1, 2, sharex=ax0)
plt.plot(tw.s, tw.betx, label='betx')
plt.plot(tw.s, tw.bety, label='bety')
# plt.plot(tw0.s, tw0.betx, '.', color='C0')
# plt.plot(tw0.s, tw0.bety, '.', color='C1')

plt.ylabel(r'$\beta_{x,y}$ [m]')
plt.legend()

ax2 = plt.subplot(3, 1, 3, sharex=ax1)
ax2.plot(tw.s, tw.dx, color='C2', label='dx')
ax2.set_ylabel(r'$D_x$ [m]')
plt.xlabel('s [m]')
plt.suptitle('Optics')

sv = line_sliced.survey()
plt.figure(2, figsize=(6.4 * 1.7, 4.8))

axsv1 = plt.subplot(1, 2, 1)
plt.plot(sv.s, sv.X, '-', label='X')
plt.plot(sv.s, sv.Z, '-', label='Z')
plt.xlabel('s [m]')
plt.ylabel('X, Z [m]')
plt.legend()
axsv2 = plt.subplot(1, 2, 2)
plt.plot(sv.Z, sv.X, '-')
plt.xlabel('Z [m]')
plt.ylabel('X [m]')
plt.axis('equal')
plt.suptitle('Survey')



tt = line.get_table(attr=True)
for ax in [ax1, ax2, axsv1]:
    ttbends = tt.rows['mb.*']
    for ii in range(len(ttbends)):
        ss = ttbends.s[ii]
        ll = ttbends.length[ii]
        ax.axvspan(ss, ss+ll, facecolor='b', alpha=0.2)

    ttquads = tt.rows['mq.*']
    for ii in range(len(ttquads)):
        ss = ttquads.s[ii]
        ll = ttquads.length[ii]
        ax.axvspan(ss, ss+ll, facecolor='r', alpha=0.2)


# Comparison of bend models
line_sliced.configure_bend_model(core='expanded', edge='linear')
tw_core_expand_fringe_linear = line_sliced.twiss(method='4d')
line_sliced.configure_bend_model(core='full', edge='linear')
tw_core_full_fringe_linear = line_sliced.twiss(method='4d')

plt.figure(3, figsize=(6.4, 4.8 * 1.4))
ax0 = plt.subplot(3, 1, 1)
ax0.plot(tw_core_expand_fringe_linear.s, tw_core_expand_fringe_linear.betx, label='betx')
ax0.plot(tw_core_full_fringe_linear.s, tw_core_full_fringe_linear.betx, label='betx')
ax0.set_ylabel(r'$\beta_{x}$ [m]')

# wx_chrom
ax1 = plt.subplot(3, 1, 2, sharex=ax0)
ax1.plot(tw_core_expand_fringe_linear.s, tw_core_expand_fringe_linear.wx_chrom, label='expanded')
ax1.plot(tw_core_full_fringe_linear.s, tw_core_full_fringe_linear.wx_chrom, label='full')
plt.ylabel(r'$W_x$')
plt.legend()

# wy_chrom
ax2 = plt.subplot(3, 1, 3, sharex=ax0)
ax2.plot(tw_core_expand_fringe_linear.s, tw_core_expand_fringe_linear.wy_chrom, label='expanded')
ax2.plot(tw_core_full_fringe_linear.s, tw_core_full_fringe_linear.wy_chrom, label='full')
plt.ylabel(r'$W_y$')


plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import matplotlib.pyplot as plt
import numpy as np
import pickle
from scipy.constants import e,c

from LHC import LHC

import xtrack as xt
xt.enable_pyheadtail_interface()

macroparticlenumber_track = 50000
macroparticlenumber_optics = 2000000
n_turns = 512

epsn_x  = 2.5e-6
epsn_y  = 3.5e-6
sigma_z = 0.05

intensity = 1e11

mode = 'smooth'
#mode = 'non-smooth'

if mode == 'smooth':
    machine = LHC(machine_configuration='Injection', n_segments=29, D_x=10)
elif mode == 'non-smooth':
    with open('lhc_2015_80cm_optics.pkl') as fid:
        optics = pickle.load(fid)
    optics.pop('circumference')

    machine = LHC(machine_configuration='Injection', optics_mode = 'non-smooth', V_RF=10e6,  **optics)

print('Create bunch for optics...')
bunch  = xt.Particles.from_pyheadtail(
        machine.generate_6D_Gaussian_bunch_matched(
        macroparticlenumber_optics, intensity, epsn_x, epsn_y, sigma_z=sigma_z))
print('Done.')



bunch.x += 10.
bunch.y += 20.
bunch.z += .020

ix = machine.one_turn_map.index(machine.longitudinal_map)
machine.one_turn_map.remove(machine.longitudinal_map)

beam_alpha_x = []
beam_beta_x = []
beam_alpha_y = []
beam_beta_y = []
for i_ele, m in enumerate(machine.one_turn_map):
    print('Element %d/%d'%(i_ele, len(machine.one_turn_map)))
    beam_alpha_x.append(bunch.alpha_Twiss_x())
    beam_beta_x.append(bunch.beta_Twiss_x())
    beam_alpha_y.append(bunch.alpha_Twiss_y())
    beam_beta_y.append(bunch.beta_Twiss_y())
    m.track(bunch)


plt.close('all')

fig, axes = plt.subplots(2, sharex=True)

axes[0].plot(np.array(beam_beta_x), 'bo')
axes[0].plot(machine.transverse_map.beta_x, 'b-', label='x')
axes[0].plot(np.array(beam_beta_y), 'ro')
axes[0].plot(machine.transverse_map.beta_y, 'r-', label='y')
axes[0].grid('on')
axes[0].set_ylabel('beta_x, beta_y')
axes[0].legend(bbox_to_anchor=(1, 1),loc='upper left',prop={'size':12})
plt.subplots_adjust(right=.86)

axes[1].plot(np.array(beam_alpha_x), 'bo')
axes[1].plot(machine.transverse_map.alpha_x, 'b-')
axes[1].plot(np.array(beam_alpha_y), 'ro')
axes[1].plot(machine.transverse_map.alpha_y, 'r-')
axes[1].grid('on')
axes[1].set_ylabel('alpha_x, alpha_y')
axes[1].set_xlabel('# point')

if mode == 'non-smooth':
    axes[0].plot(np.array(optics['beta_x']), 'xk')
    axes[0].plot(np.array(optics['beta_y']), 'xk')
    axes[1].plot(np.array(optics['alpha_x']), 'xk')
    axes[1].plot(np.array(optics['alpha_y']), 'xk')

plt.show()


machine.one_turn_map.insert(ix, machine.longitudinal_map)
bunch  = xt.Particles.from_pyheadtail(
    machine.generate_6D_Gaussian_bunch_matched(
        macroparticlenumber_track, intensity, epsn_x, epsn_y, sigma_z=sigma_z))

beam_x = []
beam_y = []
beam_z = []
sx, sy, sz = [], [], []
epsx, epsy, epsz = [], [], []
for i_turn in range(n_turns):
    print('Turn %d/%d'%(i_turn, n_turns))
    machine.track(bunch)

    beam_x.append(bunch.mean_x())
    beam_y.append(bunch.mean_y())
    beam_z.append(bunch.mean_z())
    sx.append(bunch.sigma_x())
    sy.append(bunch.sigma_y())
    sz.append(bunch.sigma_z())
    epsx.append(bunch.epsn_x()*1e6)
    epsy.append(bunch.epsn_y()*1e6)
    epsz.append(bunch.epsn_z())

plt.figure(2, figsize=(16, 8), tight_layout=True)
plt.subplot(2,3,1)
plt.plot(beam_x)
plt.ylabel('x [m]');plt.xlabel('Turn')
plt.gca().ticklabel_format(style='sci', scilimits=(0,0),axis='y')
plt.subplot(2,3,2)
plt.plot(beam_y)
plt.ylabel('y [m]');plt.xlabel('Turn')
plt.gca().ticklabel_format(style='sci', scilimits=(0,0),axis='y')
plt.subplot(2,3,3)
plt.plot(beam_z)
plt.ylabel('z [m]');plt.xlabel('Turn')
plt.gca().ticklabel_format(style='sci', scilimits=(0,0),axis='y')
plt.subplot(2,3,4)
plt.plot(np.fft.rfftfreq(len(beam_x), d=1.), np.abs(np.fft.rfft(beam_x)))
plt.ylabel('Amplitude');plt.xlabel('Qx')
plt.subplot(2,3,5)
plt.plot(np.fft.rfftfreq(len(beam_y), d=1.), np.abs(np.fft.rfft(beam_y)))
plt.ylabel('Amplitude');plt.xlabel('Qy')
plt.subplot(2,3,6)
plt.plot(np.fft.rfftfreq(len(beam_z), d=1.), np.abs(np.fft.rfft(beam_z)))
plt.xlim(0, 0.1)
plt.ylabel('Amplitude');plt.xlabel('Qz')

fig, axes = plt.subplots(3, figsize=(16, 8), tight_layout=True)
twax = [plt.twinx(ax) for ax in axes]
axes[0].plot(sx, label=r'$\sigma_x$' )
twax[0].plot(epsx, '-g', label=r'$\varepsilon_x$')
axes[0].set_xlabel('Turns')
axes[0].set_ylabel(r'$\sigma_x$')
twax[0].set_ylabel(r'$\varepsilon_x$')
axes[1].plot(sy, label=r'$\sigma_y$' )
twax[1].plot(epsy, '-g', label=r'$\varepsilon_y$')
axes[1].set_xlabel('Turns')
axes[1].set_ylabel(r'$\sigma_y$')
twax[1].set_ylabel(r'$\varepsilon_y$')
axes[2].plot(sz, label=r'$\sigma_z$' )
twax[2].plot(epsz, '-g', label=r'$\varepsilon_z$')
axes[2].set_xlabel('Turns')
axes[2].set_ylabel(r'$\sigma_z$')
twax[2].set_ylabel(r'$\varepsilon_z$')
axes[0].grid()
axes[1].grid()
axes[2].grid()
for ax in list(axes)+list(twax):
    ax.ticklabel_format(useOffset=False, style='sci', scilimits=(0,0),axis='y')
for ax in list(axes):
    ax.legend(loc='upper right',prop={'size':12})
for ax in list(twax):
    ax.legend(loc='lower right',prop={'size':12})

LHC_with_octupole_injection = LHC(machine_configuration='Injection', n_segments=5, octupole_knob = -1.5)
print('450GeV:')
print('i_octupole_focusing =',LHC_with_octupole_injection.i_octupole_focusing)
print('i_octupole_defocusing =',LHC_with_octupole_injection.i_octupole_defocusing)
print('in the machine we get 19.557')
print('  ')
LHC_with_octupole_flattop = LHC(machine_configuration='Injection', n_segments=5, p0=6.5e12*e/c, octupole_knob = -2.9)

print('6.5TeV:')
print('i_octupole_focusing =',LHC_with_octupole_flattop.i_octupole_focusing)
print('i_octupole_defocusing =',LHC_with_octupole_flattop.i_octupole_defocusing)
print('in the machine we get 546.146')

plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

# Build line with half a cell
half_cell = xt.Line(
    elements={
        'start_cell': xt.Marker(),
        'drift0': xt.Drift(length=1.),
        'qf1': xt.Quadrupole(k1=0.027/2, length=1.),
        'drift1_1': xt.Drift(length=1),
        'bend1': xt.Bend(k0=3e-4, h=3e-4, length=45.),
        'drift1_2': xt.Drift(length=1.),
        'qd1': xt.Quadrupole(k1=-0.0271/2, length=1.),
        'drift2': xt.Drift(length=1),
        'mid_cell': xt.Marker(),
    }
)
half_cell.particle_ref = xt.Particles(p0c=2e9)

# Add observation points every 1 m (to see betas inside bends)
half_cell.discard_tracker()
s_cut = np.arange(0, half_cell.get_length(), 1.)
half_cell.cut_at_s(s_cut)

# Attach knobs to quadrupoles
half_cell.vars['kqf'] = 0.027/2
half_cell.vars['kqd'] = -0.0271/2
half_cell.element_refs['qf1'].k1 = half_cell.vars['kqf']
half_cell.element_refs['qd1'].k1 = half_cell.vars['kqd']

# Match with periodic symmetric boundary
opt_halfcell = half_cell.match(
    method='4d',
    start='start_cell', end='mid_cell',
    init='periodic_symmetric',
    targets=xt.TargetSet(mux=0.2501/2, muy=0.2502/2, at='mid_cell'),
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
)

# Twiss with periodic symmetric boundary
tw_half_cell = half_cell.twiss4d(init='periodic_symmetric')

# Plot
import matplotlib.pyplot as plt
plt.close('all')
tw_half_cell.plot()
plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import logging

import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt
import xobjects as xo


plt.close('all')

n_part=10000
shift_x = 0.3e-2
shift_y = 0.5e-2

ctx = xo.context_default
buf = ctx.new_buffer()

logger = logging.getLogger('xtrack')
logger.setLevel(logging.DEBUG)

# Define aper_0
aper_0 = xt.LimitEllipse(_buffer=buf, a=2e-2, b=2e-2)
shift_aper_0 = (shift_x, shift_y)
rot_deg_aper_0 = 10.

# Define aper_1
aper_1 = xt.LimitEllipse(_buffer=buf, a=1e-2, b=1e-2)
shift_aper_1 = (shift_x, shift_y)
rot_deg_aper_1 = 10.

# aper_0_sandwitch
line_aper_0 = xt.Line(
    elements=[xt.XYShift(_buffer=buf, dx=shift_aper_0[0], dy=shift_aper_0[1]),
              xt.SRotation(_buffer=buf, angle=rot_deg_aper_0),
              aper_0,
              xt.Multipole(_buffer=buf, knl=[0.00]),
              xt.SRotation(_buffer=buf, angle=-rot_deg_aper_0),
              xt.XYShift(_buffer=buf, dx=-shift_aper_0[0], dy=-shift_aper_0[1])])
line_aper_0.build_tracker(_buffer=buf)

# aper_1_sandwitch
line_aper_1 = xt.Line(
    elements=[xt.XYShift(_buffer=buf, dx=shift_aper_1[0], dy=shift_aper_1[1]),
              xt.SRotation(_buffer=buf, angle=rot_deg_aper_1),
              aper_1,
              xt.Multipole(_buffer=buf, knl=[0.00]),
              xt.SRotation(_buffer=buf, angle=-rot_deg_aper_1),
              xt.XYShift(_buffer=buf, dx=-shift_aper_1[0], dy=-shift_aper_1[1])])
line_aper_1.build_tracker(_buffer=buf)

# Build example line
line=xt.Line(
    elements = ((xt.Drift(_buffer=buf, length=0.5),)
                + line_aper_0.elements
                + (xt.Drift(_buffer=buf, length=1),
                   xt.Multipole(_buffer=buf, knl=[0.]),
                   xt.Drift(_buffer=buf, length=1),
                   xt.Cavity(_buffer=buf, voltage=3e6, frequency=400e6),
                   xt.Drift(_buffer=buf, length=1.),)
                + line_aper_1.elements))
num_elements = len(line)

line.build_tracker()

# Test on full line
r = np.linspace(0, 0.018, n_part)
theta = np.linspace(0, 8*np.pi, n_part)
particles = xt.Particles(_context=ctx,
        p0c=6500e9,
        x=r*np.cos(theta)+shift_x,
        y=r*np.sin(theta)+shift_y)

line.track(particles)


loss_loc_refinement = xt.LossLocationRefinement(line,
                                            n_theta = 360,
                                            r_max = 0.5, # m
                                            dr = 50e-6,
                                            ds = 0.1,
                                            save_refine_lines=True,
                                            allowed_backtrack_types=[
                                                xt.Multipole, xt.Cavity])

import time
t0 = time.time()

loss_loc_refinement.refine_loss_location(particles)

t1 = time.time()
print(f'Took\t{(t1-t0)*1e3:.2f} ms')


# Automatic checks
mask_lost = particles.state == 0
r_calc = np.sqrt((particles.x-shift_x)**2 + (particles.y-shift_y)**2)
assert np.all(r_calc[~mask_lost]<1e-2)
assert np.all(r_calc[mask_lost]>1e-2)
i_aper_1 = line.elements.index(aper_1)
assert np.all(particles.at_element[mask_lost]==i_aper_1)
assert np.all(particles.at_element[~mask_lost]==0)
s0 = line.get_s_elements()[line.elements.index(aper_0)]
s1 = line.get_s_elements()[line.elements.index(aper_1)]
r0 = np.sqrt(aper_0.a_squ)
r1 = np.sqrt(aper_1.a_squ)
s_expected = s0 + (r_calc-r0)/(r1 - r0)*(s1 - s0)
# TODO This threshold is a bit large
assert np.allclose(particles.s[mask_lost], s_expected[mask_lost], atol=0.11)

# Visualize apertures
interp_line = loss_loc_refinement.refine_lines[
                                    loss_loc_refinement.i_apertures[1]]
s0 = interp_line.s0
s1 = interp_line.s1
polygon_0 = interp_line.elements[0]
polygon_1 = interp_line.elements[-1]
for ii, (trkr, poly) in enumerate(
                         zip([line_aper_0,line_aper_1],
                             [polygon_0, polygon_1])):
    part_gen_range = 0.05
    pp = xt.Particles(
                    p0c=6500e9,
                    x=np.random.uniform(-part_gen_range, part_gen_range, n_part),
                    y=np.random.uniform(-part_gen_range, part_gen_range, n_part))
    x0 = pp.x.copy()
    y0 = pp.y.copy()

    trkr.track(pp)
    ids = pp.particle_id


    fig = plt.figure(ii+1)
    ax = fig.add_subplot(111)
    plt.plot(x0, y0, '.', color='red')
    plt.axis('equal')
    plt.grid(linestyle=':')
    plt.plot(x0[ids][pp.state>0], y0[ids][pp.state>0], '.', color='green')
    plt.plot(poly.x_closed, poly.y_closed, '-k', linewidth=1)

plt.figure()
ax = plt.axes(projection='3d')
ax.plot3D(
        polygon_0.x_closed,
        polygon_0.y_closed,
        s0+polygon_0.x_closed*0,
        color='k', linewidth=3)
ax.plot3D(
        polygon_1.x_closed,
        polygon_1.y_closed,
        s1+polygon_1.x_closed*0,
        color='k', linewidth=3)
s_check = []
r_check = []
for ee, ss in zip(interp_line.elements,
                  interp_line.get_s_elements()):
    if ee.__class__ is xt.LimitPolygon:
        ax.plot3D(
                ee.x_closed,
                ee.y_closed,
                s0+ss+0*ee.x_closed,
                alpha=0.9,
                )
        s_check.append(ss+s0)
        r_vertices = np.sqrt(
                (ee.x_vertices-shift_x)**2 + (ee.y_vertices-shift_y)**2)
        r_check.append(np.mean(r_vertices))
mask = particles.at_element == loss_loc_refinement.i_apertures[1]
ax.plot3D(particles.x[mask], particles.y[mask], particles.s[mask],
          '.r', markersize=.5)
ax.view_init(65, 62); plt.draw()
plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)

p_test = line.build_particles(x=0)
line['mbw.a6r7.b1'].track(p_test)

print(p_test.x, p_test.px)


import xtrack as xt

# Create an environment
env = xt.Environment()

# Create a line with two quadrupoles and a marker
line = env.new_line(name='myline', components=[
    env.new('q0', xt.Quadrupole, length=2.0, at=10.),
    env.new('q1', xt.Quadrupole, length=2.0, at=20.),
    env.new('m0', xt.Marker, at=40.),
    ])

tt0 = line.get_table()
tt0.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# _end_point            40            40            40

# Create a new element
env.new('s1', xt.Sextupole, length=2, k2=0.2)

# Replace `q1` with `s1`
line.replace('q1', 's1')

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# s1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# _end_point            40            40            40


import numpy as np
import xtrack as xt
import xobjects as xo

env = xt.Environment.from_json('lhc.json')

assert env.lhcb1.element_dict is env.element_dict
assert env.lhcb2.element_dict is env.element_dict

assert 'drift_1' in env.lhcb1.element_names
assert 'drift_1' not in env.lhcb2.element_names

assert 'ip1' in env.lhcb1.element_names
assert 'ip1' in env.lhcb2.element_names
assert 'ip1/b2' not in env.lhcb2.element_names

assert 'mqxfa.a1r1/lhcb1' in env.lhcb1.element_names
assert 'mqxfa.a1r1/lhcb2' in env.lhcb2.element_names
assert 'mqxfa.a1r1'  not in env.lhcb1.element_names
assert 'mqxfa.a1r1'  not in env.lhcb2.element_names

assert 'mq.12l6.b1' in env.lhcb1.element_names
assert 'mq.12l6.b2' in env.lhcb2.element_names
assert 'mq.12l6.b2/lhcb2' not in env.lhcb2.element_names

twb1 = env.lhcb1.twiss4d()
twb2 = env.lhcb2.twiss4d()

assert np.all(twb1.rows['ip.*'].name ==
    np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))
assert np.all(twb2.rows['ip.*'].name ==
    np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))

xo.assert_allclose(twb1.qx, 62.31, rtol=0, atol=1e-5)
xo.assert_allclose(twb1.qy, 60.32, rtol=0, atol=1e-5)
xo.assert_allclose(twb2.qx, 62.31, rtol=0, atol=1e-5)
xo.assert_allclose(twb2.qy, 60.32, rtol=0, atol=1e-5)

env['on_disp'] = 0
env['on_x1'] = 10
env['on_x5'] = 12

twb1cross = env.lhcb1.twiss4d()
twb2cross = env.lhcb2.twiss4d()

xo.assert_allclose(twb1cross['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twb2cross['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twb1cross['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twb2cross['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

# Check combined twiss
twcross = env.twiss(method='4d')
xo.assert_allclose(twcross.lhcb1['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twcross.lhcb2['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twcross.lhcb1['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
xo.assert_allclose(twcross.lhcb2['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

env.discard_trackers()
assert env.lhcb1.tracker is None
assert env.lhcb2.tracker is None

mycontext = xo.ContextCpu()
env.build_trackers(_context=mycontext)
assert env.lhcb1.tracker is not None
assert env.lhcb2.tracker is not None
assert env.lhcb1.tracker._context is mycontext
assert env.lhcb2.tracker._context is mycontext


import xobjects as xo

from ..general import _pkg_root
from ..base_element import BeamElement
from .elements import (
    SynchrotronRadiationRecord, Bend, Quadrupole, Sextupole,
    Octupole, Solenoid, Drift, RBend, COMMON_MAGNET_SOURCES,
)
from ..random import RandomUniformAccurate, RandomExponential

from .slice_elements import _slice_copy, ID_RADIATION_FROM_PARENT


COMMON_SLICE_XO_FIELDS = {
    'radiation_flag': xo.Field(xo.Int64, default=ID_RADIATION_FROM_PARENT),
    'delta_taper': xo.Float64,
    'weight': xo.Float64,
}


class ThickSliceBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Bend), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_bend.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class ThickSliceRBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(RBend), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_rbend.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class ThickSliceQuadrupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Quadrupole), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_quadrupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class ThickSliceSextupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Sextupole), **COMMON_SLICE_XO_FIELDS}

    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_sextupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class ThickSliceOctupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Octupole), **COMMON_SLICE_XO_FIELDS}

    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord

    _extra_c_sources = [
        *COMMON_MAGNET_SOURCES,
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_octupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class ThickSliceSolenoid(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Solenoid), **COMMON_SLICE_XO_FIELDS}

    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('headers/synrad_spectrum.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thick_slice_solenoid.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj


class DriftSliceBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Bend), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice_bend.h')]

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    copy = _slice_copy

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


class DriftSliceRBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(RBend), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice_rbend.h')]

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    copy = _slice_copy

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


class DriftSliceQuadrupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Quadrupole), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice_quadrupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


class DriftSliceSextupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Sextupole), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice_sextupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


class DriftSliceOctupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Octupole), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice_octupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


class DriftSlice(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = False
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    allow_loss_refinement = True
    _force_moveable = True
    isthick = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Drift), **COMMON_SLICE_XO_FIELDS}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/drift.h'),
        _pkg_root.joinpath('beam_elements/elements_src/drift_slice.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        out = Drift(length=self._parent.length * self.weight,
                     _buffer=self._buffer)
        return out


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt
import xobjects as xo

from cpymad.madx import Madx

# Load sequence in MAD-X
mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use(sequence="lhcb1")

# Build Xtrack line importing MAD-X expressions
line = xt.Line.from_madx_sequence(mad.sequence['lhcb1'],
                                  deferred_expressions=True # <--
                                  )
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1,
                                 gamma0=mad.sequence.lhcb1.beam.gamma)
line.build_tracker()

# MAD-X variables can be found in in the imported line. They can be
# used to change properties in the beamline.
# For example, we consider the MAD-X variable `on_x1` that controls
# the beam angle in the interaction point 1 (IP1). It is defined in
# microrad units.

# Inspect the value of the variable
line['on_x1']
# returns 1 (as defined in the import)

# Measure vertical angle at the interaction point 1 (IP1)
line.twiss()['px', 'ip1']
# ---> returns 1e-6


# Set crossing angle using the variable
line['on_x1'] = 100

# Measure vertical angle at the interaction point 1 (IP1)
print(line.twiss(at_elements=['ip1'])['px'])
# ---> returns 100e-6

#!end-doc-part

#########################################################################
# The expressions relating the beam elements properties to the          #
# variables can be inspected and modified through line                  #
#########################################################################

# For example we can see how the dipole corrector 'mcbyv.4r1.b1' is controlled,
# by inspecting the expression of its normal dipole component knl[0]
line['mcbxfah.3r1'].get_expr('knl', 0)
# ---> returns "(-vars['acbxh3.r1'])"

# We can see that the variable controlling the corrector is in turn controlled
# by an expression involving several other variables:
line.get_expr('acbxh3.r1')
# ---> returns
#         (((((((-3.529000650090648e-07*vars['on_x1hs'])
#          -(1.349958221397232e-07*vars['on_x1hl']))
#          +(1.154711348310621e-05*vars['on_sep1h']))
#          +(1.535247516521591e-05*vars['on_o1h']))
#          -(9.919546388675102e-07*vars['on_a1h']))
#          +(3.769003853335184e-05*vars['on_ccpr1h']))
#          +(1.197587664190056e-05*vars['on_ccmr1h']))

# The _info() method can be used to get on overview of the information related
# to a given variable:
line.info('acbxh3.r1')
# ---> prints:
#          #  vars['acbxh3.r1']._get_value()
#             vars['acbxh3.r1'] = 0.00010587001950271944
#
#          #  vars['acbxh3.r1']._expr
#             vars['acbxh3.r1'] = (((((((-3.529000650090648e-07*vars['on_x1hs'])
#                                 -(1.349958221397232e-07*vars['on_x1hl']))
#                                 +(1.154711348310621e-05*vars['on_sep1h']))
#                                 +(1.535247516521591e-05*vars['on_o1h']))
#                                 -(9.919546388675102e-07*vars['on_a1h']))
#                                 +(3.769003853335184e-05*vars['on_ccpr1h']))
#                                 +(1.197587664190056e-05*vars['on_ccmr1h']))
#
#          #  vars['acbxh3.r1']._expr._get_dependencies()
#             vars['on_x1hs'] = -300.0
#             vars['on_a1h'] = -0.0
#             vars['on_x1hl'] = -0.0
#             vars['on_ccpr1h'] = 0.0
#             vars['on_sep1h'] = -0.0
#             vars['on_o1h'] = 0.0
#             vars['on_ccmr1h'] = 0.0
#
#          #  vars['acbxh3.r1']._find_dependant_targets()
#             element_refs['mcbxfah.3r1'].knl[0]


#########################################################################
# The Xtrack line including the related expressions can be saved in a   #
# json and reloaded.                                                    #
#########################################################################

# Save
import json
with open('status.json', 'w') as fid:
    json.dump(line.to_dict(), fid,
    cls=xo.JEncoder)

# Reload
with open('status.json', 'r') as fid:
    dct = json.load(fid)
line_reloaded = xt.Line.from_dict(dct)

import numpy as np
line.vars['on_x1'] = 250
assert np.isclose(line.twiss(at_elements=['ip1'])['px'][0], 250e-6,
                  atol=1e-6, rtol=0)

line.vars['on_x1'] = -300
assert np.isclose(line.twiss(at_elements=['ip1'])['px'][0], -300e-6,
                  atol=1e-6, rtol=0)



from cpymad.madx import Madx
import xtrack as xt
import xpart as xp
import xobjects as xo

import json

# Import a thick sequence
mad = Madx()
mad.call('../../test_data/clic_dr/sequence.madx')
mad.use('ring')

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')

# Build xtrack line
print('Build xtrack line...')
line = xt.Line.from_madx_sequence(mad.sequence['RING'])
line.particle_ref = xp.Particles(
        mass0=xp.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=mad.sequence.ring.beam.gamma
        )

with open('line.json', 'w') as f:
    json.dump(line.to_dict(), f, cls=xo.JEncoder)

# make a line suitable to test tapering

line.particle_ref = xp.Particles(
        mass0=xp.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=15000 # I push up the energy loss
        )
#line0 = line.copy()

line['rf'].voltage *= 20 # I push up the voltage to be able to compensate the energy loss

# I cycle to the center of the undulators straight section (where I want to impose delta = 0)
line = line.cycle('qdw1..1:38')

c0 = line['rf']
v0 = c0.voltage
s0 = line.get_s_position('rf')

# I install some more cavities to test the tapering
line.insert_element(at_s=41., element=c0.copy(), name='rf1')
line.insert_element(at_s=line.get_length()-s0, element=c0.copy(), name='rf2a')
line.insert_element(at_s=line.get_length()-s0, element=c0.copy(), name='rf2b')
line.insert_element(at_s=line.get_length()-41, element=c0.copy(), name='rf3')
line.insert_element(at_s=line.get_length()-41, element=c0.copy(), name='rf_off')

line['rf2a'].voltage *= 0.6 # I split the voltage unevenly to test the partitioning
line['rf2b'].voltage *= 0.4
line['rf_off'].voltage *= 0.0

with open('line_for_taper.json', 'w') as f:
    json.dump(line.to_dict(), f, cls=xo.JEncoder)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import e as qe
from scipy.constants import c as clight
from scipy.constants import epsilon_0, hbar

import xtrack as xt
import xobjects as xo

context = xo.ContextCpu()

L_bend = 1.
B_T = 1
n_part = 1_000_000

delta = 0
particles_ave = xt.Particles(
        _context=context,
        p0c=80e9, # / (1 + delta), # 5 GeV
        x=np.zeros(n_part),
        px=1e-4,
        py=-1e-4,
        delta=delta,
        mass0=xt.ELECTRON_MASS_EV)
particles_ave_0 = particles_ave.copy()
gamma = (particles_ave.energy/particles_ave.mass0)[0]
gamma0 = (particles_ave.gamma0[0])
energy = particles_ave.energy[0]
particles_rnd = particles_ave.copy()

P0_J = particles_ave.p0c[0] / clight * qe
h_bend = B_T * qe / P0_J
theta_bend = h_bend * L_bend

dipole_ave = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                          radiation_flag=1, _context=context)
dipole_rnd = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                          radiation_flag=2, _context=context)


dct_ave_before = particles_ave.to_dict()
dct_rng_before = particles_rnd.to_dict()

particles_ave._init_random_number_generator()
particles_rnd._init_random_number_generator()

dipole_ave.track(particles_ave)
dipole_rnd.track(particles_rnd)

dct_ave = particles_ave.to_dict()
dct_rng = particles_rnd.to_dict()

assert np.allclose(dct_ave['delta'], np.mean(dct_rng['delta']),
                  atol=0, rtol=5e-3)

rho_0 = L_bend/theta_bend
mass0_kg = (dct_ave['mass0']*qe/clight**2)
r0 = qe**2/(4*np.pi*epsilon_0*mass0_kg*clight**2)
Ps = (2 * r0 * clight * mass0_kg * clight**2 * gamma0**2 * gamma**2)/(3*rho_0**2) # W

Delta_E_eV = -Ps*(L_bend/clight) / qe
Delta_E_trk = (dct_ave['ptau']-dct_ave_before['ptau'])*dct_ave['p0c']

xo.assert_allclose(Delta_E_eV, Delta_E_trk, atol=0, rtol=1e-3)

# Check photons
line=xt.Line(elements=[
             xt.Drift(length=1.0),
             xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend),
             xt.Drift(length=1.0),
             xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend)
            ])
line.build_tracker(_context=context)
line.configure_radiation(model='quantum')

record_capacity = int(1000e6)
record = line.start_internal_logging_for_elements_of_type(xt.Multipole,
                                                            capacity=record_capacity)
particles_test = particles_ave_0.copy()
particles_test_before = particles_test.copy()
line.track(particles_test)

Delta_E_test = (particles_test.ptau - particles_test_before.ptau
                                                      )*particles_test.p0c
n_recorded = record._index.num_recorded
assert n_recorded < record_capacity
assert np.allclose(-np.sum(Delta_E_test), np.sum(record.photon_energy[:n_recorded]),
                  atol=0, rtol=1e-6)

p0_J = particles_ave.p0c[0] / clight * qe
B_T = p0_J / qe / rho_0
mass_0_kg = particles_ave.mass0 * qe / clight**2
E_crit_J = 3 * qe * hbar * gamma**2 * B_T / (2 * mass_0_kg)
E_crit_eV = E_crit_J / qe

E_ave_J = 8 * np.sqrt(3) / 45 * E_crit_J
E_ave_eV = E_ave_J / qe

E_sq_ave_J = 11 / 27 * E_crit_J**2
E_sq_ave_eV = E_sq_ave_J / qe**2

assert np.isclose(np.mean(record.photon_energy[:n_recorded]), E_ave_eV, rtol=1e-2, atol=0)
assert np.isclose(np.std(record.photon_energy[:n_recorded]), np.sqrt(E_sq_ave_eV - E_ave_eV**2), rtol=1e-3, atol=0)

hist, bin_edges = np.histogram(np.log10(record.photon_energy[:n_recorded]),
                range = (np.log10(1e-4 * E_crit_eV),
                         np.log10(1e4 * E_crit_eV)), bins=500)
bin_centers = (bin_edges[:-1] + bin_edges[1:])/2

dE = np.diff(10**bin_edges)
E_center = 10**bin_centers

dn_dE = hist / dE / (2 * L_bend / clight) / n_part

xo.assert_allclose(np.sum(dn_dE * dE * E_center) * qe, Ps, atol=0, rtol=1e-2)

dn_dE_at_E_crit = np.interp(E_crit_eV, E_center, dn_dE)

dn_dE_norm = dn_dE / dn_dE_at_E_crit

# Check against analytical expresson from:
# Implements from A. Hofmann, The physics of synchrotron radiation, Eq. 5.48b
import n_photons
dn_dE_ref = n_photons.photon_spectrum(E_center, gamma, 1/h_bend)
dn_dE_ref_at_E_crit = np.interp(E_crit_eV, E_center, dn_dE_ref)
dn_dE_ref_norm = dn_dE_ref / dn_dE_ref_at_E_crit

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.loglog(E_center/E_crit_eV, dn_dE_norm, '.', label='Xsuite - photon histogram')
plt.loglog(E_center/E_crit_eV, dn_dE_ref_norm, '-', label='Analytic')
plt.legend()
plt.xlabel(r'$E/E{_\text{crit}}$')
plt.ylabel('Normalized dN/dE')
plt.xlim(1e-3, 1e1)
plt.ylim(1e-5, 1e3)
plt.grid(True)

plt.figure(2)
plt.loglog(E_center, dn_dE, label='Xsuite photon histogram')
plt.loglog(E_center, dn_dE_ref, '--', label='Analytical')
plt.axvline(x=E_crit_eV, linestyle='--', color='k', label='Critical energy')
plt.xlim(1e-3*E_crit_eV, 1e1*E_crit_eV)
plt.ylim(1.e-2, 1e5)
plt.xlabel('Energy [eV]')
plt.ylabel('dN/dE [1/(eV s)]')
plt.legend()

plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=-10.0, from_='q0'),
        env.new('qr', 'Quadrupole', length=2.0, at=10.0, from_='q0'),
        env.new('mk1', 'Marker', at=40),
        env.new('mk2', 'Marker', at=42),
        env.new('end', 'Marker', at=50.),
    ])

s_tol = 1e-10

tt0 = line.get_table()
tt0.show(cols=['name', 's_start', 's_end', 's_center'])

env.new('ss', 'Sextupole', length='0.1')
line.insert([
    env.new('q1', 'q0', at=-5.0, from_='ql'),
    env.place('ss'),
    env.new('q2', 'q0', anchor='start', at=15.0 - 1.),
    env.place('ss'),
    env.new('q3', 'q0', anchor='start', at=29, from_='ql@end'),
    env.place('ss'),
])

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1..0', 'q1', 'ss::0', 'drift_1..3', 'ql', 'drift_2..0',
       'q2', 'ss::1', 'drift_2..3', 'q0', 'drift_3', 'qr',
       'drift_4', 'mk1', 'q3', 'mk2', 'ss::2', 'drift_6..1', 'end',
       '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 2.  ,  5.  ,  6.05,  7.55, 10.  , 12.5 , 15.  , 16.05, 17.55,
      20.  , 25.  , 30.  , 35.5 , 40.  , 41.  , 42.  , 42.05, 46.05,
      50.  , 50.  ]), rtol=0., atol=1e-14)



import xtrack as xt
import xobjects as xo
from xtrack.beam_elements.magnets import Magnet

env = xt.Environment()
env.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV, energy0=100e9)

line = env.new_line(components=[
    env.new('mb', 'Bend', angle=0.01, length=10., k0_from_h=True),
])

line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(100))
    ]
)

line.build_tracker()
line.configure_radiation(model='mean')

p_ref = env.particle_ref.copy()
line.track(p_ref)

mm = Magnet(angle=0.01, length=10., k0_from_h=True)
mm.radiation_flag = 1 # 1: mean
mm.num_multipole_kicks = 100
mm.integrator = 'uniform'
p_test = line.particle_ref.copy()

mm.track(p_test)

xo.assert_allclose(p_ref.delta, p_test.delta, atol=0, rtol=2e-4)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time

import numpy as np
from matplotlib import pyplot as plt
from scipy import constants

import xobjects as xo
context = xo.ContextCpu(omp_num_threads=0)
import xtrack as xt

muonMass = constants.value('muon mass energy equivalent in MeV')*1E6

n_macroparticles = 10
energy = 1.25 # [GeV]
p0c = np.sqrt((energy*1E9)**2-muonMass**2)
gamma = np.sqrt(1+(p0c/muonMass)**2)
beta = np.sqrt(1-1/gamma**2)
normemit_x = 25E-6
normemit_y = 25E-6
sigma_z = 4.6E-2
emit_z = 7.5E-3
sigma_delta = emit_z/sigma_z/energy
beta_s = sigma_z/sigma_delta

beta_x_IP1 = 10.0
beta_y_IP1 = 10.0
alpha_x_IP1 = 0.0
alpha_y_IP1 = 0.0
dispersion_IP1 = 0.0

beta_x_IP2 = 0.1
beta_y_IP2 = 0.1
alpha_x_IP2 = 0.0
alpha_y_IP2 = 2.0
dispersion_IP2 = 1.0

nTurn = 500
energy_increment = (5.0-1.25)*1E9/nTurn/2
#energy_increment = 0.0

particles = xt.Particles(_context=context,
                         q0 = 1,
                         mass0 = muonMass,
                         p0c=p0c,
                         x=np.sqrt(normemit_x*beta_x_IP1/gamma)*np.linspace(0,6,n_macroparticles),
                         px=np.sqrt(normemit_x/beta_x_IP1/gamma)*np.zeros(n_macroparticles,dtype=float),
                         y=np.sqrt(normemit_y*beta_y_IP1/gamma)*np.linspace(0,6,n_macroparticles),
                         py=np.sqrt(normemit_y/beta_y_IP1/gamma)*np.zeros(n_macroparticles,dtype=float),
                         zeta=sigma_z*np.zeros(n_macroparticles),
                         delta=sigma_delta*np.ones(n_macroparticles,dtype=float),
                         )

arc12 = xt.LineSegmentMap(
        alfx=[alpha_x_IP1,alpha_x_IP2],
        betx=[beta_x_IP1,beta_x_IP2],
        dx=[dispersion_IP1,dispersion_IP2],
        alfy=[alpha_y_IP1,alpha_y_IP2],
        bety=[beta_y_IP1,beta_y_IP2],
        qx=0.155, qy=0.16, qs=0.155, bets=-beta_s, #Careful with this sign!
        energy_ref_increment=energy_increment,
        energy_increment=0)

arc21 = xt.LineSegmentMap(
        alfx=[alpha_x_IP2,alpha_x_IP1],
        betx=[beta_x_IP2,beta_x_IP1],
        dx=[dispersion_IP2,dispersion_IP1],
        alfy=[alpha_y_IP2,alpha_y_IP1],
        bety=[beta_y_IP2,beta_y_IP1],
        qx=0.155, qy=0.16, qs=0.155, bets=-beta_s, #Careful with this sign!
        energy_ref_increment=energy_increment,
        energy_increment=0)

plt.figure(12)
plt.plot(particles.zeta,particles.delta,'.b')
plt.axvline(sigma_z,color='k')
plt.axvline(-sigma_z,color='k')
plt.axhline(sigma_delta,color='k')
plt.axhline(-sigma_delta,color='k')


for turn in range(nTurn):
    plt.figure(0)
    plt.plot(particles.x,particles.px,'.b')
    plt.figure(1)
    plt.plot(particles.y,particles.py,'.g')
    plt.figure(3)
    plt.plot(turn,0.5*((particles.x[1]-dispersion_IP1*particles.delta[1])**2*(1+alpha_x_IP1**2)/beta_x_IP1+2*alpha_x_IP1*(particles.x[1]-dispersion_IP1*particles.delta[1])*particles.px[1]+particles.px[1]**2*beta_x_IP1)*particles.gamma0[1]*particles.beta0[1],'x')
    arc12.track(particles)
    plt.figure(10)
    plt.plot(particles.x,particles.px,'.b')
    plt.figure(11)
    plt.plot(particles.y,particles.py,'.g')
    plt.figure(13)
    plt.plot(turn,0.5*((particles.x[1]-dispersion_IP2*particles.delta[1])**2*(1+alpha_x_IP2**2)/beta_x_IP2+2*alpha_x_IP2*(particles.x[1]-dispersion_IP2*particles.delta[1])*particles.px[1]+particles.px[1]**2*beta_x_IP2)*particles.gamma0[1]*particles.beta0[1],'x')
    arc21.track(particles)

    plt.figure(12)
    plt.plot(particles.zeta,particles.delta,'.r')

    plt.figure(100)
    plt.plot(turn,particles.mass0*particles.gamma0[1]*1E-9,'xk')

plt.show()






import numpy as np
from cpymad.madx import Madx
import xtrack as xt

mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use(sequence='lhcb1')
tw_mad = mad.twiss()
summ_mad = mad.table.summ
surv_mad = mad.survey()

line = xt.Line.from_madx_sequence(mad.sequence['lhcb1'],
                                     deferred_expressions=True)
line.build_tracker()
surv_xt = line.survey()

import matplotlib.pyplot as plt
plt.figure(1)
plt.plot(surv_mad.s, surv_mad.X, 'b')
plt.plot(surv_xt.s, surv_xt.X, '--', color='lightblue')
plt.plot(surv_mad.s, surv_mad.Z, 'r')
plt.plot(surv_xt.s, surv_xt.Z, '--', color='darkred')

plt.figure(2)
plt.plot(surv_mad.s, surv_mad.phi, 'b')
plt.plot(surv_xt.s, surv_xt.phi, '--', color='lightblue')
plt.plot(surv_mad.s, surv_mad.theta, 'r')
plt.plot(surv_xt.s, surv_xt.theta, '--', color='darkred')
plt.plot(surv_mad.s, surv_mad.psi, 'g')
plt.plot(surv_xt.s, surv_xt.psi, '--', color='darkgreen')

plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

context = xo.ContextCpu()

n_part=300000

particles = xt.Particles(_context=context,
        p0c=6500e9,
        x=np.random.uniform(-0.25, 0.25, n_part),
        px = np.zeros(n_part),
        y=np.random.uniform(0, 0.1, n_part),
        py = np.zeros(n_part))

tilt_deg = 10.
aper = xt.LimitRect(_context=context,
                    min_x=-.1,
                    max_x=.1,
                    min_y=-0.001,
                    max_y=0.001,
                    shift_x=0.08,
                    shift_y=0.04,
                    rot_s_rad=np.deg2rad(tilt_deg))

aper.track(particles)

part_id = context.nparray_from_context_array(particles.particle_id)
part_state = context.nparray_from_context_array(particles.state)
part_x = context.nparray_from_context_array(particles.x)
part_y = context.nparray_from_context_array(particles.y)

x_alive = part_x[part_state>0]
y_alive = part_y[part_state>0]

assert_allclose = np.testing.assert_allclose
assert_allclose(np.mean(x_alive), 0.08, rtol=5e-2, atol=0)
assert_allclose(np.mean(y_alive), 0.04, rtol=5e-2, atol=0)
slope = np.polyfit(x_alive, y_alive, 1)[0]
assert_allclose(slope, np.tan(np.deg2rad(tilt_deg)), rtol=5e-2, atol=0)

# Correlation coefficient


id_alive = part_id[part_state>0]

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(part_x, part_y, '.', color='red')
plt.plot(part_x[part_state>0], part_y[part_state>0], '.', color='green')
plt.axis('equal')

plt.show()



import json

import numpy as np
import xtrack as xt
import xobjects as xo

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct)

line.build_tracker()

tw_before = line.twiss()

GreaterThan = xt.GreaterThan
LessThan = xt.LessThan

tw0 = line.twiss()
opt = line.match(
    solve=False,
    solver='jacobian',
    # Portion of the beam line to be modified and initial conditions
    start='mq.33l8.b1',
    end='mq.17l8.b1',
    init=tw0, init_at=xt.START,
    # Dipole corrector strengths to be varied
    vary=[
        xt.Vary(name='acbv28.l8b1', step=1e-10),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
        xt.Vary(name='acbv22.l8b1', step=1e-10),
    ],
    targets=[
        xt.Target('y', GreaterThan(2.7e-3), at='mb.b26l8.b1'),
        xt.Target('y', GreaterThan(2.7e-3), at='mb.b25l8.b1'),
        xt.Target('y', at='mq.24l8.b1', value=xt.LessThan(3e-3)),
        xt.Target('y', at='mq.26l8.b1', value=xt.LessThan(6e-3)),
        xt.TargetSet(['y', 'py'], at='mq.17l8.b1', value=tw0),
    ]
)


# Check freeze
opt.step(1)
ts = opt.target_status(ret=True)
assert ts.tol_met[2] == False
assert ts.tol_met[4] == False
assert ts.tol_met[5] == False

opt.targets[2].freeze()
opt.targets[5].freeze()

ts = opt.target_status(ret=True)
assert ts.tol_met[2] == True
assert ts.tol_met[4] == False
assert ts.tol_met[5] == True

opt.targets[2].unfreeze()
opt.targets[5].unfreeze()

ts = opt.target_status(ret=True)
assert ts.tol_met[2] == False
assert ts.tol_met[4] == False
assert ts.tol_met[5] == False

opt.solve()

tw = line.twiss()
assert tw['y', 'mb.b26l8.b1'] > 2.7e-3
assert tw['y', 'mb.b25l8.b1'] > 2.7e-3
assert tw['y', 'mq.24l8.b1'] < 3e-3 + 1e-6
assert np.isclose(tw['y', 'mq.17l8.b1'], tw_before['y', 'mq.17l8.b1'], rtol=0, atol=1e-7)
assert np.isclose(tw['py', 'mq.17l8.b1'], tw_before['py', 'mq.17l8.b1'], rtol=0, atol=1e-9)

assert isinstance(opt.targets[0].value, xt.GreaterThan)
assert isinstance(opt.targets[1].value, xt.GreaterThan)
assert isinstance(opt.targets[2].value, xt.LessThan)

assert opt.targets[0].value._value == 0
assert opt.targets[1].value._value == 0
assert opt.targets[2].value._value == 0

assert opt.targets[0].value.mode == 'step'
assert opt.targets[1].value.mode == 'step'
assert opt.targets[2].value.mode == 'step'

assert opt.targets[0].value.lower == 2.7e-3
assert opt.targets[1].value.lower == 2.7e-3
assert opt.targets[2].value.upper == 3e-3


# Test mode smooth
# Remove the bump

for kk in ['acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1', 'acbv22.l8b1']:
    line.vars[kk] = 0

tw_before = line.twiss()
assert tw_before['y', 'mb.b26l8.b1'] < 1e-7
assert tw_before['y', 'mb.b25l8.b1'] < 1e-7

opt = line.match(
    solve=False,
    solver='jacobian',
    # Portion of the beam line to be modified and initial conditions
    start='mq.33l8.b1',
    end='mq.17l8.b1',
    init=tw_before, init_at=xt.START,
    # Dipole corrector strengths to be varied
    vary=[
        xt.Vary(name='acbv28.l8b1', step=1e-10),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
        xt.Vary(name='acbv22.l8b1', step=1e-10),
    ],
    targets=[
        xt.Target('y', GreaterThan(2.7e-3, mode='smooth', sigma_rel=0.05), at='mb.b26l8.b1'),
        xt.Target('y', GreaterThan(2.7e-3, mode='smooth'), at='mb.b25l8.b1'),
        xt.Target('y', at='mq.24l8.b1', value=xt.LessThan(3e-3, mode='smooth', sigma_rel=0.04)),
        xt.Target('y', at='mq.26l8.b1', value=xt.LessThan(6e-3, mode='smooth')),
        xt.TargetSet(['y', 'py'], at='mq.17l8.b1', value=tw_before),
    ]
)

# Check freeze
opt.step(1)
ts = opt.target_status(ret=True)
assert ts.tol_met[0] == False
assert ts.tol_met[1] == False
assert ts.tol_met[5] == False

opt.targets[0].freeze()
opt.targets[5].freeze()

ts = opt.target_status(ret=True)
assert ts.tol_met[0] == True
assert ts.tol_met[1] == False
assert ts.tol_met[5] == True

opt.targets[0].unfreeze()
opt.targets[5].unfreeze()

ts = opt.target_status(ret=True)
assert ts.tol_met[0] == False
assert ts.tol_met[1] == False
assert ts.tol_met[5] == False

opt.solve()

tw = line.twiss()

assert tw['y', 'mb.b26l8.b1'] > 2.7e-3 - 1e-6
assert tw['y', 'mb.b25l8.b1'] > 2.7e-3 - 1e-6
assert tw['y', 'mq.24l8.b1'] < 3e-3 + 1e-6
assert tw['y', 'mq.26l8.b1'] < 6e-3 + 1e-6
assert np.isclose(tw['y', 'mq.17l8.b1'], tw_before['y', 'mq.17l8.b1'], rtol=0, atol=1e-7)
assert np.isclose(tw['py', 'mq.17l8.b1'], tw_before['py', 'mq.17l8.b1'], rtol=0, atol=1e-9)

assert isinstance(opt.targets[0].value, xt.GreaterThan)
assert isinstance(opt.targets[1].value, xt.GreaterThan)
assert isinstance(opt.targets[2].value, xt.LessThan)
assert isinstance(opt.targets[3].value, xt.LessThan)

assert opt.targets[0].value._value == 0
assert opt.targets[1].value._value == 0
assert opt.targets[2].value._value == 0
assert opt.targets[3].value._value == 0

assert opt.targets[0].value.mode == 'smooth'
assert opt.targets[1].value.mode == 'smooth'
assert opt.targets[2].value.mode == 'smooth'
assert opt.targets[3].value.mode == 'smooth'

assert opt.targets[0].value.lower == 2.7e-3
assert opt.targets[1].value.lower == 2.7e-3
assert opt.targets[2].value.upper == 3e-3
assert opt.targets[3].value.upper == 6e-3

assert np.isclose(opt.targets[0].value.sigma, 0.05 * 2.7e-3, atol=0, rtol=1e-10)
assert np.isclose(opt.targets[1].value.sigma, 0.01 * 2.7e-3, atol=0, rtol=1e-10)
assert np.isclose(opt.targets[2].value.sigma, 0.04 * 3e-3, atol=0, rtol=1e-10)
assert np.isclose(opt.targets[3].value.sigma, 0.01 * 6e-3, atol=0, rtol=1e-10)

x_cut_norm = 1/16 + np.sqrt(33)/16
poly = lambda x: 3 * x**3 - 2 * x**4

# Check smooth target (GreaterThan)
i_tar_gt = 0
tar_gt = opt.targets[i_tar_gt]
sigma_gt = tar_gt.value.sigma
x0_gt = tar_gt.runeval()

edge_test_gt = np.linspace(x0_gt - 3 * sigma_gt, x0_gt + 3 * sigma_gt, 100)

residue_gt = edge_test_gt * 0
for ii, xx in enumerate(edge_test_gt):
    tar_gt.value.lower = xx
    residue_gt[ii] =  opt._err()[i_tar_gt] / tar_gt.weight

x_minus_edge_gt = x0_gt - edge_test_gt
x_cut_gt = -x_cut_norm * sigma_gt

mask_zero_gt = x_minus_edge_gt > 0
assert np.all(residue_gt[mask_zero_gt] == 0)
mask_linear_gt = x_minus_edge_gt < x_cut_gt
assert np.allclose(residue_gt[mask_linear_gt],
    -x_minus_edge_gt[mask_linear_gt] - x_cut_norm * sigma_gt + sigma_gt*poly(x_cut_norm),
    atol=0, rtol=1e-10)
mask_poly_gt = (~mask_zero_gt) & (~mask_linear_gt)
assert np.allclose(residue_gt[mask_poly_gt],
    sigma_gt * poly(-x_minus_edge_gt[mask_poly_gt]/sigma_gt),
    atol=0, rtol=1e-10)

# Check smooth target (LessThan)
i_tar_lt = 2
tar_lt = opt.targets[i_tar_lt]
sigma_lt = tar_lt.value.sigma
x0_lt = tar_lt.runeval()

edge_test_lt = np.linspace(x0_lt - 3 * sigma_lt, x0_lt + 3 * sigma_lt, 100)

residue_lt = edge_test_lt * 0
for ii, xx in enumerate(edge_test_lt):
    tar_lt.value.upper = xx
    residue_lt[ii] =  opt._err()[i_tar_lt] / tar_lt.weight

x_minus_edge_lt = x0_lt - edge_test_lt
x_cut_lt = x_cut_norm * sigma_lt

mask_zero_lt = x_minus_edge_lt < 0
assert np.all(residue_lt[mask_zero_lt] == 0)
mask_linear_lt = x_minus_edge_lt > x_cut_lt
assert np.allclose(residue_lt[mask_linear_lt],
    x_minus_edge_lt[mask_linear_lt] - x_cut_norm * sigma_lt + sigma_lt*poly(x_cut_norm),
    atol=0, rtol=1e-10)
mask_poly_lt = (~mask_zero_lt) & (~mask_linear_lt)
assert np.allclose(residue_lt[mask_poly_lt],
    sigma_lt * poly(x_minus_edge_lt[mask_poly_lt]/sigma_lt),
    atol=0, rtol=1e-10)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(100)
plt.suptitle('GreaterThan')
plt.plot(x_minus_edge_gt, residue_gt)
plt.plot(x_minus_edge_gt,
    -x_minus_edge_gt - x_cut_norm * sigma_gt + sigma_gt*poly(x_cut_norm),
    '--', color='C1')
plt.plot(x_minus_edge_gt, sigma_gt * poly(-x_minus_edge_gt/sigma_gt), '--', color='C1')
plt.axvline(x=-x_cut_gt, color='r', linestyle='--')
plt.axvline(x=0, color='r', linestyle='--')
plt.axvline(x=-sigma_gt, color='g', linestyle='--')
plt.ylim(np.array([-1, 1]) * np.max(np.abs(x_minus_edge_gt)))

plt.figure(101)
plt.suptitle('LessThan')
plt.plot(x_minus_edge_lt, residue_lt)
plt.plot(x_minus_edge_lt,
    x_minus_edge_lt - x_cut_norm * sigma_lt + sigma_lt*poly(x_cut_norm),
    '--', color='C1')
plt.plot(x_minus_edge_lt, sigma_lt * poly(x_minus_edge_lt/sigma_lt), '--', color='C1')
plt.axvline(x=x_cut_lt, color='r', linestyle='--')
plt.axvline(x=0, color='r', linestyle='--')
plt.axvline(x=sigma_lt, color='g', linestyle='--')
plt.ylim(np.array([-1, 1]) * np.max(np.abs(x_minus_edge_lt)))



fig = plt.figure(1, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw_before.s, tw_before.y*1000, label='y')
ax.plot(tw.s, tw.y*1000, label='y')
# Target
ax.axvline(x=line.get_s_position('mb.b26l8.b1'), color='r', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mb.b25l8.b1'), color='r', linestyle='--', alpha=0.5)
# Correctors
ax.axvline(x=line.get_s_position('mcbv.32l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.28l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.26l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.24l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.22l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.18l8.b1'), color='k', linestyle='--', alpha=0.5)
# Boundaries
ax.axvline(x=line.get_s_position('mq.33l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.17l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.set_xlim(line.get_s_position('mq.33l8.b1') - 10,
            line.get_s_position('mq.17l8.b1') + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-10, 10)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)
plt.show()



import xtrack as xt
import xobjects as xo

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=7e12)

class MyKicker:
    def __init__(self, kickx=0, kicky=0):
        self.kickx = kickx
        self.kicky = kicky

    def track(self, p):
        p.px += self.kickx
        p.py += self.kicky

kick1 = MyKicker(kickx=1e-3)
kick2 = MyKicker(kickx=-2e-3)
kick3 = MyKicker(kickx=1e-3)

line = env.new_line(components=[
    env.place('k1', kick1, at=0),
    env.place('k2', kick2, at=2),
    env.place('k3', kick3, at=4),
    ])

tw = line.twiss(betx=1, bety=1, include_collective=True)
xo.assert_allclose(tw['x', 'k2'], 2e-3, atol=1e-8, rtol=0)

tw_nocoll = line.twiss(betx=1, bety=1)
xo.assert_allclose(tw_nocoll['x', 'k2'], 0, atol=1e-8, rtol=0)


class MyQuad:
    def __init__(self, k1l=0):
        self.k1l = k1l

    def track(self, p):
        p.px += -self.k1l * p.x
        p.py += self.k1l * p.y

kq = 0.1
qf = MyQuad(k1l=0.1)
qd = MyQuad(k1l=-0.1)

lfodo = env.new_line(components=[
    env.place('qf', qf, at=0),
    env.place('qd', qd, at=10),
    env.new('end', 'Marker', at=20),
])

twfodo = lfodo.twiss4d(include_collective=True)

lfodo_ref = env.new_line(components=[
    env.new('mqf', 'Multipole', knl=[0, kq], at=0),
    env.new('mqd', 'Multipole', knl=[0, -kq], at=10),
    env.new('mend', 'Marker', at=20),
])

twfodo_ref = lfodo_ref.twiss4d(include_collective=True)

xo.assert_allclose(twfodo.qx, twfodo_ref.qx, atol=1e-8, rtol=0)
xo.assert_allclose(twfodo.qy, twfodo_ref.qy, atol=1e-8, rtol=0)
xo.assert_allclose(twfodo.dqx, twfodo_ref.dqx, atol=1e-8, rtol=0)
xo.assert_allclose(twfodo.dqy, twfodo_ref.dqy, atol=1e-8, rtol=0)

twfodo_open = lfodo.twiss4d(include_collective=True, betx=1, bety=1,
                            compute_chromatic_properties=True)
twfodo_open_ref = lfodo_ref.twiss4d(include_collective=True, betx=1, bety=1,
                                    compute_chromatic_properties=True)

xo.assert_allclose(twfodo_open.betx, twfodo_open_ref.betx, atol=0, rtol=1e-8)
xo.assert_allclose(twfodo_open.bety, twfodo_open_ref.bety, atol=0, rtol=1e-8)
xo.assert_allclose(twfodo_open.ax_chrom, twfodo_open_ref.ax_chrom, atol=0, rtol=1e-8)
xo.assert_allclose(twfodo_open.ay_chrom, twfodo_open_ref.ay_chrom, atol=0, rtol=1e-8)


import numpy as np
import pandas as pd
import xtrack as xt

line = xt.Line.from_json('psb_04_with_chicane_corrected_thin.json')
line.build_tracker()
line.t_turn_s = 0 # Reset time!

line.vars['on_chicane_k0'] = 1
line.vars['on_chicane_k2'] = 1
line.vars['on_chicane_beta_corr'] = 1
line.vars['on_chicane_tune_corr'] = 1

# Install monitor at foil
monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=6000, num_particles=1)
line.discard_tracker()
line.insert_element(index='bi1.tstr1l1', element=monitor, name='monitor_at_foil')
line.build_tracker()

fname = '../../test_data/psb_chicane/inj_distrib.dat'
df = pd.read_table(fname, skiprows=3,
    names="x x' y y' z z' Phase Time Energy Loss".split())

kin_energy_ev = df.Energy.values * 1e6
tot_energy_ev = kin_energy_ev + xt.PROTON_MASS_EV
p0c = line.particle_ref.p0c[0]
tot_energy0_ev = line.particle_ref.energy0[0]

ptau = (tot_energy_ev - tot_energy0_ev) / p0c


part_for_injection = xt.Particles(q0=1, mass0=xt.PROTON_MASS_EV, p0c=line.particle_ref.p0c[0],
                                  ptau=ptau)

part_for_injection.x = df.x.values * 1e-3
part_for_injection.y = df.y.values * 1e-3
part_for_injection.zeta = df.z.values * 1e-3
part_for_injection.px = df["x'"].values  * 1e-3 * (1 + part_for_injection.delta)
part_for_injection.py = df["y'"].values  * 1e-3 * (1 + part_for_injection.delta)
part_for_injection.weight = 10




p_injection = xt.ParticlesInjectionSample(particles_to_inject=part_for_injection,
                                        line=line,
                                        element_name='injection',
                                        num_particles_to_inject=7)

line.discard_tracker()
line.insert_element(index='bi1.tstr1l1', element=p_injection, name='injection')


# Add monitor at end line
monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=10, num_particles=100)
line.insert_element(index='p16ring1$end', element=monitor, name='monitor_at_end')

line.build_tracker()

p = line.build_particles(_capacity=100, x=0)
p.state[0] = -500 # kill the particle added by default

intensity = []
line.enable_time_dependent_vars = True
for iturn in range(8):
    intensity.append(p.weight[p.state>0].sum())
    line.track(p, num_turns=1)

assert np.all(np.sum(monitor.state > 0, axis=0)
              == np.array([ 0,  7, 14, 21, 28, 35, 42, 49, 56,  0]))

assert np.all(monitor.at_element[monitor.state > 0] ==
              line.element_names.index('monitor_at_end'))

assert np.allclose(monitor.s[monitor.state > 0],
                   line.get_s_position('monitor_at_end'), atol=1e-12)